#include "prefix.h"

#include "image.h"
#include "exception.h"
#include "error.h"
#include "safeinteger.h"
#include "ucharbuffer.h"

#include <windows.h>
#include <assert.h>
#include <limits.h>

/*************************************************

	摜obt@

**************************************************/

APP_BEGIN_NAMESPACE

class ImageMember
{
public:
	ImageMember()
	{
	}
	ImageMember(const ImageMember &source)
		:bitmapInfo(source.bitmapInfo)
		,buffer(source.buffer)
		,size(source.size)
		,pixelBytes(source.pixelBytes)
	{
	}
	~ImageMember()
	{
	}
	BITMAPINFO_FIELDS bitmapInfo;
	UCharBuffer buffer;
	SIZE size;
	int pixelBytes;
};

Image::Image()
:m_m(new ImageMember)

{
	memset(&(m_m->bitmapInfo),0,sizeof(BITMAPINFO_FIELDS));
	m_m->bitmapInfo.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
	m_m->bitmapInfo.bmiHeader.biPlanes=1;
}

Image::Image(const Image &source)
:m_m(new ImageMember(*(source.m_m)))
{
}

Image::~Image() throw()
{
	Free();
}

const Image &Image::operator=(const Image &source)
{
	*m_m=*source.m_m;
	return *this;
}

void Image::Dup(const Image &source)
{
	if(Width()!=source.Width()
		|| Height()!=source.Height()
		|| PixelBytes()!=source.PixelBytes()) {
		SIZE size;
		size.cx=source.Width();
		size.cy=source.Height();
		Create(size,source.PixelBytes());
	}
	memcpy(GetW(),source.Get(),TotalBytes());
}

//obt@p
void Image::Free() throw()
{
	m_m->buffer.Free();
}

//
int Image::Width() const
{
	return m_m->size.cx;
}

//c
int Image::Height() const
{
	return m_m->size.cy;
}

//s̃oCg
int Image::RowBytes() const
{
	bool success;
	int rowBytes=SafeAddition(RowDataBytes(),PadBytes(),&success);
	if(!success) {
		throw Exception(APPERR_INVALID_IMAGE_SIZE);
	}
	return rowBytes;
}

//s̋lߕ̃oCg
int Image::PadBytes() const
{
	return (-RowDataBytes())&(sizeof(LONG)-1);
}

//s̓ẽoCg
int Image::RowDataBytes() const
{
	bool success;
	int rowDataBytes=SafeMultiple(Width(),PixelBytes(),&success);
	if(!success) {
		throw Exception(APPERR_INVALID_IMAGE_SIZE);
	}
	return rowDataBytes;
}

//obt@̃oCg
int Image::TotalBytes() const
{
	bool success;
	int totalBytes=SafeMultiple(RowBytes(),Height(),&success);
	if(!success) {
		throw Exception(APPERR_INVALID_IMAGE_SIZE);
	}
	return totalBytes;
}

//obt@mۂ
void Image::Create(SIZE size,int pixelBytes)
{
	Free();
	Resize(size,pixelBytes);
}

//obt@̂܂܂ɂđ傫ς
//eʂȂꍇ₷
void Image::Resize(SIZE size,int pixelBytes)
{
	if(size.cx<0 || size.cy<0) {
		throw Exception(APPERR_INVALID_IMAGE_SIZE);
	}

	assert(1<=pixelBytes && pixelBytes<=5);

	int oldTotalBytes;
	if(m_m->buffer.IsEmpty()) {
		oldTotalBytes=0;
	} else {
		oldTotalBytes=TotalBytes();
	}

	m_m->pixelBytes=pixelBytes;
	m_m->size=size;
	m_m->buffer.Increase(TotalBytes());

	if(m_m->pixelBytes==3) {
		m_m->bitmapInfo.bmiHeader.biBitCount=24;
		m_m->bitmapInfo.bmiHeader.biWidth=size.cx;
		m_m->bitmapInfo.bmiHeader.biHeight=-size.cy;
		m_m->bitmapInfo.bmiHeader.biCompression=BI_RGB;
	}
}

//obt@̃AhX
unsigned char *Image::GetW()
{
	return m_m->buffer.GetW();
}

const unsigned char *Image::Get() const
{
	return m_m->buffer.Get();
}

//w肵s̃AhX
unsigned char *Image::GetRowW(int y)
{
	assert(0<=y && 0<m_m->size.cy);
	return m_m->buffer.GetW()+RowBytes()*y;
}

const unsigned char *Image::GetRow(int y) const
{
	assert(0<=y && 0<m_m->size.cy);
	return m_m->buffer.Get()+RowBytes()*y;
}

//w肵W̃AhX
unsigned char *Image::GetPixelW(int x,int y)
{
	assert(0<=y && 0<m_m->size.cy);
	return m_m->buffer.GetW()+RowBytes()*y+PixelBytes()*x;
}

const unsigned char *Image::GetPixel(int x,int y) const
{
	assert(0<=y && 0<m_m->size.cy);
	return m_m->buffer.Get()+RowBytes()*y+PixelBytes()*x;
}

//EBhEỸrbg}bv
const BITMAPINFO *Image::BitmapInfo() const
{
	assert(m_m->pixelBytes==3);
	return reinterpret_cast<const BITMAPINFO*>(&m_m->bitmapInfo);
}

//Pf̃oCg
int Image::PixelBytes() const
{
	return m_m->pixelBytes;
}

//摜̑傫
SIZE Image::Size() const
{
	return m_m->size;
}

//摜Sď
void Image::Clear()
{
	memset(GetW(),0,TotalBytes());
}

APP_END_NAMESPACE
